-- Beispiele aus "Introduction to Functional Programming using Haskell"
--

module BinTree where

-- 1. Binre Bume

data BTree a = Leaf a 
	     | Fork (BTree a) (BTree a)
	       deriving (Show)


-- einfache Operationen auf Binrbumen


-- Size
size :: BTree a -> Int
size (Leaf x) = 1
size (Fork xt yt) = size xt + size yt

-- Flatten

flatten :: BTree a -> [a]
flatten (Leaf x) = [x]
flatten (Fork xt yt) = flatten xt ++ flatten yt

-- Nodes

nodes :: BTree a -> Int
nodes (Leaf x) = 0
nodes (Fork xt yt) = 1 + nodes xt + nodes yt

-- Height

height :: BTree a -> Int
height (Leaf x) = 0
height (Fork xt yt) = 1 + max (height xt) (height yt)

-- showTree

showBTree :: (Show a ) => BTree a -> String
showBTree (Leaf x) = show x
showBTree (Fork xt yt) = "(" ++ showBTree xt ++ "," ++ showBTree yt ++ ")"

-- depths

depths :: BTree a -> BTree Int
depths = down 0

down :: Int -> BTree a -> BTree Int
down n (Leaf f) = Leaf n
down n (Fork xt yt ) = Fork (down (n+1) xt) (down (n+1) yt)

-- maxBtree

maxBtree :: (Ord a) => BTree a -> a
maxBtree (Leaf f) = f
maxBtree (Fork xt yt) = max (maxBtree xt) (maxBtree yt)

-- erzeugen von Binrbumen aus Listen

mkBTree :: [a] -> BTree a
mkBTree xs
    | (m == 0) = Leaf (unwrap xs) -- ein-elementige Liste
    | otherwise = Fork (mkBTree ys) (mkBTree zs)
    where m = (length xs) `div` 2
          (ys, zs) = splitAt m xs
          unwrap [x] = x

-- Map & Fold

mapBTree :: (a -> b) -> BTree a -> BTree b
mapBTree f (Leaf x) = Leaf (f x)
mapBTree f (Fork xt yt) = Fork (mapBTree f xt) (mapBTree f yt)

foldBTree :: (a -> b) -> (b -> b -> b) -> BTree a -> b
foldBTree f g (Leaf x) = f x
foldBTree f g (Fork xt yt) = g (foldBTree f g xt) (foldBTree f g yt)

-- Retrieve (Spezifikation)

retrieveBTree :: BTree a -> Int -> a
retrieveBTree t k = (!!) (flatten t) k

-- height und size durch fold ausdrcken

size2 = foldBTree (const 1) (+)
height2 = foldBTree (const 0) maxsubs
	  where m `maxsubs` n = 1 + m `max` n

-- Ein Beispielbaum

t1,b1,b2 :: BTree Int
t1 = mkBTree [100,1,10,20,32,0,129,28,85,58,17]

b1 = Leaf 10
b2 = Fork (Leaf 10) (Fork (Leaf 3) (Leaf 2))

{- zum Vorfhren:



flatten (Fork (Fork (Leaf 3) (Fork (Leaf 2) (Leaf 1))) (Leaf 5))
-}
