-- Beispiele aus "Introduction to Functional Programming using Haskell"
--

module SearchTree where

-- Typdefinition eines Suchbaums

data (Ord a) => Stree a = Null
			| Fork (Stree a) a (Stree a)
			  deriving(Show)

-- flatten

flatten	:: (Ord a) => Stree a -> [a]
flatten Null = []
flatten (Fork xt x yt) = flatten xt ++ [x] ++ flatten yt

-- member

member :: (Ord a) => a -> Stree a -> Bool
member x Null = False
member x (Fork xt y yt)
	   | x < y = member x xt
	   | x == y = True
	   | x > y = member x yt

-- height

height :: (Ord a) => Stree a -> Int
height Null = 0
height (Fork xt y yt) = 1 + (height xt `max` height yt)

-- mkStree

mkStree :: (Ord a) => [a] -> Stree a
mkStree [] = Null
mkStree (x:xs) = Fork (mkStree smaller) x (mkStree larger)
		 where (smaller,larger) = partition (<= x) xs

-- partition

partition :: (a->Bool) -> [a] -> ([a],[a])
partition p xs = ([ x | x <- xs, p x],
		  [ y | y <- xs, (not . p) y])

-- insert & delete

insert :: (Ord a) => a -> Stree a -> Stree a
insert x Null = Fork Null x Null
insert x (Fork xt y yt)
	   | (x < y ) = Fork (insert x xt) y yt
	   | (x == y) = Fork xt y yt
	   | (x > y) = Fork xt y (insert x yt)

delete :: (Ord a) => a -> Stree a -> Stree a
delete x Null = Null
delete x (Fork xt y yt)
	   | (x < y) = Fork (delete x xt) y yt
	   | (x == y) = join xt yt
	   | (x > y) = Fork xt y (delete x yt)

join :: (Ord a) => Stree a -> Stree a -> Stree a
join xt yt = if empty yt then xt else Fork xt (headTree yt) (tailTree yt)

empty :: (Ord a) => Stree a -> Bool
empty Null = True
empty (Fork xt y yt) = False

splitTree :: (Ord a) => Stree a -> (a,Stree a)
splitTree (Fork xt y yt) = if empty xt then (y,yt) else (x, Fork wt y yt)
			   where (x,wt) = splitTree xt

headTree :: (Ord a) => Stree a -> a
headTree = fst . splitTree

tailTree :: (Ord a) => Stree a -> Stree a
tailTree = snd . splitTree

sort :: (Ord a) => [a] -> [a]
sort = flatten . mkStree

-- ein Bespielbaum

t1 :: Stree Int
t1 = mkStree [18,15,12,8,3,2,4]

{- Beispiele:

insert 7 (mkStree [6,4,2,10])
delete 10 (mkStree [6,4,2,10])
-}
